// Arduino Nano VLF Metal Detector - Nano #2
// Sketch for 2-line 16-character LCD
// Displays target ID number (VDI), signal strength, target indicator bar, and battery icon
// LCD is cleared after approximately 3 seconds with no target signal

#include <LiquidCrystal.h>

// LCD / Arduino Nano connections
//  1 - GND
//  2 - +5V
//  3 - contrast pot
//  4 - (RS) pin 12
//  5 - (RW) pin 11
//  6 - (EN) pin 10
//  7 - NC
//  8 - NC
//  9 - NC
// 10 - NC
// 11 - (Data4) pin 9
// 12 - (Data5) pin 8
// 13 - (Data6) pin 7
// 14 - (Data7) pin 6
// 15 - (BL+) pin 13 (via resistor)
// 16 - (BL-) GND

// LCD pin assignments (RS, RW, EN, Data7, Data6, Data5, Data4)
LiquidCrystal lcd(12, 11, 10, 9, 8, 7, 6);

// Other pin assignments
byte debugPin = 4;  // For software debugging and scope sync
byte compPin = 5;   // Target trigger from comparator
byte gbPin = A0;    // ADC input for GB channel
byte discPin = A1;  // ADC input for DISC channel
byte battPin = A2;  // ADC input for battery measurement
byte amDisc = 2;    // AM/DISC switch   

// Program constants
const byte sampleDelay = 23;       // Number of milliseconds to delay target sample
const byte numSamples = 128;       // Number of samples for averaging
const word dispCountLimit = 2000;  // Display counter limit

// Program variables
float gb;            // GB channel reading
float gbAcc = 0;     // Accumulator for GB channel readings
float disc;          // DISC channel reading
float discAcc = 0;   // Accumulator for DISC channel readings
float angleRad;      // Target phase angle in radians
float angleDeg;      // Target angle in degrees
float magnitude;     // Signal magnitude
int vdi = 0;         // Target visual display indicator
int sig = 0;         // Target signal strength
byte vdiBar = 0;     // Position of cursor on VDI bar
float batt;          // Battery voltage reading
word dispCount = 0;  // Clear display counter

void checkBatt() {
  batt = analogRead(battPin);      // Read battery voltage
  batt *= 4.88E-3 * 3;             // Scale ADC reading
  // Display battery condition icons
  lcd.setCursor(15,1);             // Set cursor to column 15, row 1
  if (batt >= 12) {                // Check if battery voltage >= 12V
    lcd.write(byte(0));    
  }else{
    if (batt >= 11.7) {            // Check if battery voltage >= 11.7V
      lcd.write(byte(1));
    }else{ 
      if (batt >= 11.4) {          // Check if battery voltage >= 11.4V
        lcd.write(byte(2));
      }else{
        if (batt >= 11.1) {        // Check if battery voltage >= 11.1V
          lcd.write(byte(3));
        }else{
          if (batt >= 10.8) {      // Check if battery voltage >= 10.8V
            lcd.write(byte(4));
          }else{
            if (batt >= 10.5) {    // Check if battery voltage >= 10.5V
              lcd.write(byte(5));
            }else{
              lcd.print("X");      // Battery needs charging
            }
          }
        }
      }
    }
  }
}

void setup() {
  pinMode(debugPin, OUTPUT);      // Set debug pin as an output
  pinMode(amDisc, INPUT_PULLUP);  // Enable input resistor for AM/DISC switch
  lcd.begin(16,2);                // Define LCD as having 16 columns and 2 rows
  lcd.clear();                    // Clear LCD
  lcd.setCursor(0,0);             // Set cursor to column 0, row 0
  lcd.print("Arduino Nano VLF");  // Display line 0 title
  lcd.setCursor(0,1);             // Set cursor to column 0, row 1
  lcd.print(" Metal Detector ");  // Display line 1 title
  delay (3000);                   // Display title lines for 3 seconds
  lcd.clear();                    // Clear LCD
  lcd.setCursor(0,0);             // Set cursor to column 0, row 0
  // Set up LCD with ID, SIG markers, and VDI bar
  lcd.print("ID=");          
  lcd.setCursor(7,0);
  lcd.print("SIG=");
  lcd.setCursor(0,1);
  lcd.print("------");
  lcd.print("|");
  lcd.print("-------");
  // Load custom characters into CG RAM of LCD
  // Battery condition icons
  byte batt120V[8] = {14,31,31,31,31,31,31,31};
  lcd.createChar(0, batt120V);
  byte batt117V[8] = {14,31,17,31,31,31,31,31};
  lcd.createChar(1, batt117V);
  byte batt114V[8] = {14,31,17,17,31,31,31,31};
  lcd.createChar(2, batt114V);
  byte batt111V[8] = {14,31,17,17,17,31,31,31};
  lcd.createChar(3, batt111V);
  byte batt108V[8] = {14,31,17,17,17,17,31,31};
  lcd.createChar(4, batt108V);
  byte batt105V[8] = {14,31,17,17,17,17,17,31};
  lcd.createChar(5, batt105V);
  // Block character for VDI bar
  byte block[8] = {31,17,17,17,17,17,31,0};
  lcd.createChar(6, block);
  checkBatt();             // Check battery voltage
}

void loop() {                                       
  if (digitalRead(compPin) == LOW) {     // Check for target trigger
    delay(sampleDelay);                  // Delay ADC measurements
    gbAcc = 0;                           // Reset GB accumulator
    discAcc = 0;                         // Reset DISC accumulator
    digitalWrite(debugPin, HIGH);        // Set debug pin high
    for (int i=1; i<=numSamples; i++) {  // Take several GB and DISC readings
      gb = analogRead(gbPin);            // Read GB channel
      disc = analogRead(discPin);        // Read DISC channel
      gbAcc += gb;                       // Accumulate GB readings
      discAcc += disc;                   // Accumulate DISC readings
    }
    digitalWrite(debugPin, LOW);         // Set debug pin low
    gb = gbAcc / numSamples;             // Average GB readings
    disc = discAcc / numSamples;         // Average DISC readings
    if (gb >= 512) {                     // Re-scale GB reading
      gb -= 512;
    }
    if (disc >= 512) {                   // Re-scale DISC reading
      disc -= 512;
    } else {
      disc = (512 - disc) * -1;    
    }
    lcd.setCursor(3,0);                       // Clear ID display
    lcd.print("    ");
    lcd.setCursor(3,0);
    angleRad = atan(disc/gb);                 // Calculate target phase in radians
    angleDeg = angleRad / TWO_PI * 360;       // Convert to degrees
    vdi = int(angleDeg);                      // Round value for display
    lcd.print(vdi);                           // Display target  
    magnitude = sqrt((gb*gb)+(disc*disc));    // Calculate target signal magnitude
    sig = int(magnitude);                     // Round value for display
    lcd.setCursor(11,0);                      // Clear SIG display
    lcd.print("   ");
    lcd.setCursor(11,0);
    lcd.print(sig);                           // Display target signal strength
    if (vdiBar != 6) {                        // Do not overwrite VDI bar separator
      lcd.setCursor(vdiBar,1);                // Set cursor to previous VDI bar position
      lcd.print("-");                         // Overwrite with a "dash"
    } 
    vdiBar = vdi / 10 + 6;                    // Calculate position for VDI cursor
    if (vdiBar != 6) {                        // Do not overwrite VDI bar separator
      if (vdiBar > 13) { vdiBar = 13; }       // Do not write beyond end of VDI bar
      lcd.setCursor(vdiBar,1);                // Set cursor to new VDI bar position
      lcd.write(byte(6));                     // Display a block character
    }
    dispCount = 0;                            // Reset display counter
    delay (100);                              // Wait for 100ms before next reading
  }else{
    checkBatt();                              // Check battery voltage
    lcd.setCursor(15,0);                      // Set cursor to last character on line 1
    if (digitalRead(amDisc) == LOW){          // Check AM/DISC switch
      lcd.print("D");                         // Detector in DISC mode
    } else {
      lcd.print("A");                         // Detector in AM mode      
    }
    dispCount++;                              // Increment display counter
    if (dispCount >= dispCountLimit) {        // Check if display counter has reached limit
      dispCount = 0;                          // Reset display counter
      lcd.setCursor(3,0);                     // Clear ID, SIG, and VDI bar
      lcd.print("    ");
      lcd.setCursor(11,0);
      lcd.print("   ");
      lcd.setCursor(0,1);
      lcd.print("------");
      lcd.print("|");
      lcd.print("-------");
    }
  }
}
