// Arduino Nano VLF Metal Detector - Nano #2
// Sketch for 4-line 16-character LCD
// Displays target ID number, VDI bar, signal strength, SIG bar, and battery voltage
// LCD is cleared after approximately 3 seconds with no target signal

#include <LiquidCrystal.h>

// LCD / Arduino Nano connections
//  1 - GND
//  2 - +5V
//  3 - contrast pot
//  4 - (RS) pin 12
//  5 - (RW) pin 11
//  6 - (EN) pin 10
//  7 - NC
//  8 - NC
//  9 - NC
// 10 - NC
// 11 - (Data4) pin 9
// 12 - (Data5) pin 8
// 13 - (Data6) pin 7
// 14 - (Data7) pin 6
// 15 - (BL+) pin 13 (via resistor)
// 16 - (BL-) GND

// LCD pin assignments (RS, RW, EN, Data7, Data6, Data5, Data4)
LiquidCrystal lcd(12, 11, 10, 9, 8, 7, 6);

// Other pin assignments
byte debugPin = 4;  // For software debugging and scope sync
byte compPin = 5;   // Target trigger from comparator
byte gbPin = A0;    // ADC input for GB channel
byte discPin = A1;  // ADC input for DISC channel
byte battPin = A2;  // ADC input for battery measurement
byte amDisc = 2;    // AM/DISC switch   

// Program constants
const byte sampleDelay = 23;       // Number of milliseconds to delay target sample
const byte numSamples = 128;       // Number of samples for averaging
const word dispCountLimit = 2000;  // Display counter limit

// Program variables
float gb;            // GB channel reading
float gbAcc = 0;     // Accumulator for GB channel readings
float disc;          // DISC channel reading
float discAcc = 0;   // Accumulator for DISC channel readings
float angleRad;      // Target phase angle in radians
float angleDeg;      // Target angle in degrees
float magnitude;     // Signal magnitude
int vdi = 0;         // Target visual display indicator
int sig = 0;         // Target signal strength
byte vdiBar = 0;     // Position of cursor on VDI bar
byte sigBar = 0;     // Position of cursor on SIG bar
float batt;          // Battery voltage reading
int battV = 0;       // Battery voltage as an integer
int battDisp = 0;    // Battery voltage digit to display
word dispCount = 0;  // Clear display counter

void checkBatt() {
  batt = analogRead(battPin);   // Read battery voltage
  batt *= 4.88E-3 * 3;          // Scale ADC reading
  battV = int(batt * 10);       // Battery voltage to one decimal place
  battDisp = battV / 100 % 10;  // Extract 1st digit
  lcd.setCursor(11,2);
  lcd.print(battDisp);          // Display 1st digit
  battDisp = battV / 10 % 10;   // Extract 2nd digit
  lcd.print(battDisp);          // Display 2nd digit
  lcd.print(".");               // Display decimal place
  battDisp = battV % 10;        // Extract 3rd digit
  lcd.print(battDisp);          // Display 3rd digit
  lcd.print("V");
}

void setup() {
  pinMode(debugPin, OUTPUT);      // Set debug pin as an output
  pinMode(amDisc, INPUT_PULLUP);  // Enable pullup resistor for AM/DISC switch
  lcd.begin(16,4);                // Define LCD as having 16 columns and 4 rows
  lcd.clear();                    // Clear LCD
  lcd.setCursor(0,1);             // Set cursor to column 0, row 1
  lcd.print("Arduino Nano VLF");  // Display line 0 title
  lcd.setCursor(0,2);             // Set cursor to column 0, row 2
  lcd.print(" Metal Detector ");  // Display line 1 title
  delay (3000);                   // Display title lines for 3 seconds
  lcd.clear();                    // Clear LCD
  lcd.setCursor(0,0);             // Set cursor to column 0, row 0
  // Set up LCD with ID, signal strength, VDI bar, and SIG bar
  lcd.print("ID=");          
  lcd.setCursor(0,1);
  lcd.print("------");
  lcd.print("|");
  lcd.print("-------");
  lcd.setCursor(0,2);
  lcd.print("SIG=");
  lcd.setCursor(15,2);
  lcd.print("V");
  lcd.setCursor(0,3);
  lcd.print("--------------");
  // Load custom character into CG RAM of LCD
  // Block character for VDI and SIG bars
  byte block[8] = {31,17,17,17,17,17,31,0}; 
  lcd.createChar(0, block);
  checkBatt();                    // Check battery voltage
}

void loop() {                                       
  if (digitalRead(compPin) == LOW) {          // Check for target trigger
    delay(sampleDelay);                       // Delay ADC measurements
    gbAcc = 0;                                // Reset GB accumulator
    discAcc = 0;                              // Reset DISC accumulator
    digitalWrite(debugPin, HIGH);             // Set debug pin high
    for (int i=1; i<=numSamples; i++) {       // Take several GB and DISC readings
      gb = analogRead(gbPin);                 // Read GB channel
      disc = analogRead(discPin);             // Read DISC channel
      gbAcc += gb;                            // Accumulate GB readings
      discAcc += disc;                        // Accumulate DISC readings
    }
    digitalWrite(debugPin, LOW);              // Set debug pin low
    gb = gbAcc / numSamples;                  // Average GB readings
    disc = discAcc / numSamples;              // Average DISC readings
    if (gb >= 512) {                          // Re-scale GB reading
      gb -= 512;
    }
    if (disc >= 512) {                        // Re-scale DISC reading
      disc -= 512;
    } else {
      disc = (512 - disc) * -1;    
    }
    lcd.setCursor(3,0);                       // Clear ID display
    lcd.print("    ");
    lcd.setCursor(3,0);
    angleRad = atan(disc/gb);                 // Calculate target phase in radians
    angleDeg = angleRad / TWO_PI * 360;       // Convert to degrees
    vdi = int(angleDeg);                      // Round value for display
    lcd.print(vdi);                           // Display target  
    magnitude = sqrt((gb*gb)+(disc*disc));    // Calculate target signal magnitude
    sig = int(magnitude);                     // Round value for display
    lcd.setCursor(11,0);                      // Clear SIG display
    lcd.print("   ");
    lcd.setCursor(4,2);
    lcd.print(sig);                           // Display target signal strength
    if (vdiBar != 6) {                        // Do not overwrite VDI bar separator
      lcd.setCursor(vdiBar,1);                // Set cursor to previous VDI bar position
      lcd.print("-");                         // Overwrite with a "dash"
    } 
    vdiBar = vdi / 10 + 6;                    // Calculate correct position VDI bar indicator
    if (vdiBar != 6) {                        // Do not overwrite VDI bar separator
      if (vdiBar > 13) { vdiBar = 13; }       // Do not write beyond end of VDI bar
      lcd.setCursor(vdiBar,1);                // Set cursor to new VDI bar position
      lcd.write(byte(0));                     // Display a block character
    }
    lcd.setCursor(sigBar,3);                  // Set cursor to previous SIG bar position
    lcd.print("-");                           // Overwrite with a "dash"
    sigBar = sig / 55;                        // Calculate correct position SIG bar indicator
    if (sigBar > 13) { sigBar = 13; }         // Do not write beyond end of SIG bar
    lcd.setCursor(sigBar,3);                  // Set cursor to new SIG bar position
    lcd.write(byte(0));                       // Display a block character
    dispCount = 0;                            // Reset display counter
    delay (100);                              // Wait for 100ms before next reading
  }else{
    lcd.setCursor(11,0);                      // Set cursor to last character on line 1
    if (digitalRead(amDisc) == LOW){          // Check AM/DISC switch
      lcd.print("DISC");                      // Detector in DISC mode
    } else {
      lcd.print("  AM");                      // Detector in AM mode      
    }
    dispCount++;                              // Increment display counter
    if (dispCount >= dispCountLimit) {        // Check if display counter has reached limit
      dispCount = 0;                          // Reset display counter
      lcd.setCursor(3,0);                     // Clear ID, signal strength, VDI and SIG bars
      lcd.print("    ");
      lcd.setCursor(4,2);
      lcd.print("   ");
      lcd.setCursor(0,1);
      lcd.print("------");
      lcd.print("|");
      lcd.print("-------");
      lcd.setCursor(0,3);
      lcd.print("--------------");
      checkBatt();                            // Check battery voltage
    }
  }
}
