// Arduino Nano VLF Metal Detector - Nano #2
// Sketch for OLED
// Displays target ID number (VDI), signal strength, and target indicator bar

// Include text-only OLED library
// Writes directly to display
// No buffer in micro required
#include <U8x8lib.h>

// Constructor for AZ Delivery 1,3 Zoll I2C OLED display
U8X8_SH1106_128X64_NONAME_HW_I2C u8x8(U8X8_PIN_NONE);

// OLED / Arduino Nano connections
// VDD - +5V
// GND - GND
// SCK - A5
// SDA - A4

// Pin assignments
byte SDAPin = A4;   // Assign OLED serial data to A4
byte SCLPin = A5;   // Assign OLED serial clock to A5

// Other pin assignments
byte debugPin = 7;  // For software debugging and scope sync
byte compPin = 5;   // Target trigger from comparator
byte gbPin = A0;    // ADC input for GB channel
byte discPin = A1;  // ADC input for DISC channel
byte battPin = A2;  // ADC input for battery measurement
byte amDisc = 2;    // AM/DISC switch

// Program constants
const byte sampleDelay = 23;       // Number of milliseconds to delay target sample
const byte numSamples = 128;       // Number of samples for averaging
const word dispCountLimit = 250;   // Display counter limit

// Program variables
uint8_t block[8] = {0x7e,0x7e,0x7e,0x7e,0x7e,0x7e,0x7e,0x7e};  // Custom block character
float gb;   
float gbAcc = 0;
float disc;  
float discAcc = 0;
float angleRad;
float angleDeg;
float magnitude;
int vdi = 0;
int sig = 0;
byte vdiBar = 1;
byte sigBar = 1;
float batt;
int battV = 0;
int battDisp = 0;
word dispCount = 0;  // Clear display counter

void checkBatt() {
  batt = analogRead(battPin);   // Read battery voltage
  batt *= 4.88E-3 * 3;          // Scale ADC reading
  battV = int(batt * 10);       // Battery voltage to one decimal place
  battDisp = battV / 100 % 10;  // Extract 1st digit
  u8x8.setCursor(11,5);
  u8x8.print(battDisp);         // Display 1st digit
  battDisp = battV / 10 % 10;   // Extract 2nd digit
  u8x8.print(battDisp);         // Display 2nd digit
  u8x8.print(".");              // Display decimal place
  battDisp = battV % 10;        // Extract 3rd digit
  u8x8.print(battDisp);         // Display 3rd digit
  u8x8.print("V");
}

void setup()
{
  pinMode(debugPin, OUTPUT);                   // Set debug pin as an output
  pinMode(amDisc, INPUT_PULLUP);               // Enable input resistor for AM/DISC switch 
  u8x8.begin();                                // Initialise display
  u8x8.setFont(u8x8_font_victoriabold8_r);
  u8x8.draw1x2String(0,0,"----------------");  // Display splash screen
  u8x8.draw1x2String(0,2,"ARDUINO NANO VLF");
  u8x8.draw1x2String(0,4," METAL DETECTOR ");
  u8x8.draw1x2String(0,6,"----------------");
  delay(3000);                                 // Display title lines for 3 seconds
  // Set up OLED with ID, SIG markers, and VDI bar
  u8x8.clearDisplay();                         // Clear OLED display
  u8x8.setCursor(0,0);
  u8x8.print("ID=");
  u8x8.setCursor(0,5);
  u8x8.print("SIG=");
  u8x8.setCursor(1,2);
  u8x8.print("------|-------");                // Display middle marker (vertical line)
  u8x8.setCursor(1,7);
  u8x8.print("--------------");
  checkBatt();
}

void loop() {                                       
  if (digitalRead(compPin) == LOW) {          // Check for target trigger
    delay(sampleDelay);                       // Delay ADC measurements
    gbAcc = 0;                                // Reset GB accumulator
    discAcc = 0;                              // Reset DISC accumulator
    digitalWrite(debugPin, HIGH);             // Set debug pin high
    for (int i=1; i<=numSamples; i++) {       // Take several GB and DISC readings
      gb = analogRead(gbPin);                 // Read GB channel
      disc = analogRead(discPin);             // Read DISC channel
      gbAcc += gb;                            // Accumulate GB readings
      discAcc += disc;                        // Accumulate DISC readings
    }
    digitalWrite(debugPin, LOW);              // Set debug pin low
    gb = gbAcc / numSamples;                  // Average GB readings
    disc = discAcc / numSamples;              // Average DISC readings
    if (gb >= 512) {                          // Re-scale GB reading
      gb -= 512;
    }
    if (disc >= 512) {                        // Re-scale DISC reading
      disc -= 512;
    } else {
      disc = (512 - disc) * -1;    
    }
    u8x8.setCursor(3,0);                      // Clear ID display
    u8x8.print("     ");
    u8x8.setCursor(3,0);
    angleRad = atan(disc/gb);                 // Calculate target phase in radians
    angleDeg = angleRad / TWO_PI * 360;       // Convert to degrees
    vdi = int(angleDeg);                      // Round value for display
    u8x8.print(vdi);                          // Display target  
    magnitude = sqrt((gb*gb)+(disc*disc));    // Calculate target signal magnitude
    sig = int(magnitude);                     // Round value for display
    u8x8.setCursor(4,5);                      // Clear SIG display
    u8x8.print("     ");
    u8x8.setCursor(4,5);
    u8x8.print(sig);                          // Display target signal strength
    if (vdiBar != 7) {                        // Do not overwrite VDI bar seperator
      u8x8.setCursor(vdiBar,2);               // Set cursor to previous VDI bar position
      u8x8.print("-");                        // Overwrite with a "dash"
    }
    vdiBar = vdi / 10 + 7;                    // Calculate position for VDI cursor
    if (vdiBar != 7) {                        // Do not overwrite VDI bar separator
      if (vdiBar > 14) { vdiBar = 14; };      // Do not write beyond VDI bar position
      u8x8.setCursor(vdiBar, 2);              // Set cursor to new VDI bar position
      u8x8.drawTile(vdiBar,2,1,block);        // Display a block character                    
    }
    u8x8.setCursor(sigBar,7);                 // Set cursor to previous SIG bar position
    u8x8.print("-");                          // Overwrite with a "dash"
    sigBar = sig / 55 + 1;                    // Calculate correct position SIG bar indicator
    if (sigBar > 14) { sigBar = 14; }         // Do not write beyond end of SIG bar
    u8x8.setCursor(sigBar,7);                 // Set cursor to new SIG bar position
    u8x8.drawTile(sigBar,7,1,block);          // Display a block character
    dispCount = 0;                            // Reset display counter
    delay(100);                               // Wait for 100ms for next target indication
  } else {
    checkBatt();                              // Check battery voltage
    u8x8.setCursor(12,0);
    if (digitalRead(amDisc) == LOW){          // Check AM/DISC switch
      u8x8.print("DISC");                     // Detector in DISC mode
    } else {
      u8x8.print("AM  ");                     // Detector in AM mode      
    }
    dispCount++;                              // Increment display countervdiBagggggg
    if (dispCount >= dispCountLimit) {        // Check if display counter has reached limit
      dispCount = 0;                          // Reset display counter
      u8x8.setCursor(3,0);                    // Clear ID, SIG, and VDI bar
      u8x8.print("    ");
      u8x8.setCursor(4,5);
      u8x8.print("   ");
      u8x8.setCursor(1,2);
      u8x8.print("------");
      u8x8.print("|");
      u8x8.print("-------");
      u8x8.setCursor(1,7);
      u8x8.print("--------------");
    }
  }
}
