// Arduino Nano VLF Metal Detector - Nano #2
// Sketch for TFT display
// Displays target ID number (VDI), signal strength, and confidence bar

// Display is a Waveshare 2.4" TFT (320x240) controlled via Adafruit libraries
#include <SPI.h>
#include <Adafruit_GFX.h>
#include <Adafruit_ILI9341.h>

// TFT SPI definitions
#define TFT_MISO 12  // Master in Slave out (MISO) - not used by Waveshare TFT
#define TFT_MOSI 11  // Master out Slave in (MOSI)
#define TFT_SCK 10   // Synchronous clock signal
#define TFT_CS 9     // Chip select
#define TFT_DC 8     // Data / Control
#define TFT_RESET 7  // Reset
#define TFT_LED 6    // Backlight

// Instantiate ILI9341 TFT
Adafruit_ILI9341 tft = Adafruit_ILI9341(TFT_CS, TFT_DC, TFT_MOSI, TFT_SCK, TFT_RESET, TFT_MISO);

// Other pin assignments
byte debugPin = 4;  // For software debugging and scope sync
byte compPin = 5;   // Target trigger from comparator
byte gbPin = A0;    // ADC input for GB channel
byte discPin = A1;  // ADC input for DISC channel
byte battPin = A2;  // ADC input for battery measurement
byte amDisc = 2;    // AM/DISC switch

// Program constants
const byte sampleDelay = 23;      // Number of milliseconds to delay target sample
const byte numSamples = 128;      // Number of samples for averaging
const word dispCountLimit = 200;  // Display counter limit

//Program variables
float gb;                                 // GB channel reading
float gbAcc = 0;                          // Accumulator for GB channel readings
float disc;                               // DISC channel reading
float discAcc = 0;                        // Accumulator for DISC channel readings
float angleRad;                           // Target phase in radians
float angleDeg;                           // Target phase in degrees   
float magnitude;                          // Signal magnitude
int vdi = 0;                              // Target visual display indicator
byte vdiBar = 1;                          // Position of cursor on VDI bar
int sig = 0;                              // Signal strength
int sigBar;                               // Length of signal bar
int prevSigBar = 0;                       // Previous length of signal bar
int sigBarY;                              // Y position of cursor on SIG bar
int prevSigBarY = 194;                    // Previous Y position of cursor on SIG bar
int temp;                                 // Temorary variable
byte mode = 0;                            // Operating mode (AM or DISC)
byte prevMode = 2;                        // Previous operating mode
float batt;                               // Battery voltage reading
int battV = 0;                            // Battery voltage as an integer
int battDisp = 0;                         // Battery voltage digit to display
word dispCount = 0;                       // Clear display counter
byte backlight = 192;                     // PWM value for backlight brightness
byte targetID;                            // Target ID number
uint16_t targetIDColor;                   // Target ID colour (green or red)
uint16_t battColor;                       // Battery voltage display colour (green or red)
uint16_t prevBattColor = ILI9341_BLACK;   // Previous battery voltage colour
byte battArray[] = {99, 99, 99, 99, 99};  // Record of battery voltage digits
byte acceptBar = 0;                       // Number of accept bars for CONF display
byte rejectBar = 0;                       // Number of reject bars for CONF display

void checkBatt() {
  batt = analogRead(battPin);   // Read battery voltage
  batt *= 4.88E-3 * 3;          // Scale ADC reading
  battV = int(batt * 10);       // Battery voltage to one decimal place
  if (battV >= 105) {               
    battColor = ILI9341_GREEN;  // Display colour is green when voltage >= 10.5V
  } else {
    battColor = ILI9341_RED;    // Display colour is red when battery voltage < 10.5V
  }
  // Only update battery voltage display digits if voltage has changed
  battDisp = battV / 100 % 10;  // Extract 1st digit
  if ((battArray[0] != battDisp) | (battColor != prevBattColor)) {
    tft.drawChar(128, 214, 32, battColor, ILI9341_BLACK, 2);  // Overwrite previous value
    tft.drawChar(128, 214, battDisp + 48, battColor, ILI9341_BLACK, 2);  // Display 1st digit        
    battArray[0] = battDisp;  // Save new value for ist digit
  }
  battDisp = battV / 10 % 10;  // Extract 2nd digit
  if ((battArray[1] != battDisp) | (battColor != prevBattColor)) {
    tft.drawChar(140, 214, 32, battColor, ILI9341_BLACK, 2);  // Overwrite previous value
    tft.drawChar(140, 214, battDisp + 48, battColor, ILI9341_BLACK, 2);  // Display 2nd digit
    battArray[1] = battDisp;  // Save new value for 2nd digit
  }
  if ((battArray[2] != 46) | (battColor != prevBattColor)) {
    tft.drawChar(152, 214, 32, battColor, ILI9341_BLACK, 2);
    tft.drawChar(152, 214, 46, battColor, ILI9341_BLACK, 2);  // Display decimal place
    battArray[2] = 46; 
  }
  battDisp = battV % 10;  // Extract 3rd digit
  if ((battArray[3] != battDisp) | (battColor != prevBattColor)) {
    tft.drawChar(164, 214, 32, battColor, ILI9341_BLACK, 2);  // Overwrite previous value
    tft.drawChar(164, 214, battDisp + 48, battColor, ILI9341_BLACK, 2);  // Display 3rd digit
    battArray[3] = battDisp;  // Save new value for 3rd digit
  }
  if ((battArray[4] != 86) | (battColor != prevBattColor)) {
    tft.drawChar(176, 214, 32, battColor, ILI9341_BLACK, 2);
    tft.drawChar(176, 214, 86, battColor, ILI9341_BLACK, 2);  // Display voltage sign
    battArray[4] = 86;
  }
  prevBattColor = battColor;  // Save battery voltage colour
}

void setup()
{
  pinMode(debugPin, OUTPUT);        // Set debug pin as an output
  pinMode(amDisc, INPUT_PULLUP);    // Enable pullup resistor for AM/DISC switch
  analogWrite(TFT_LED, backlight);  // Turn on TFT backlight
  tft.begin();                      // Init display
  tft.setRotation(3);               // Rotate display to landscape with connector on LHS
  tft.fillScreen(ILI9341_BLACK);    // Clear display

  // Draw screen border
  tft.drawRect(0, 0, 320, 240, ILI9341_WHITE);    
  
  // Inner display area
  tft.drawLine(98, 239, 98, 142, ILI9341_WHITE);
  tft.drawLine(98, 142, 88, 132, ILI9341_WHITE);
  tft.drawLine(88, 132, 88, 48, ILI9341_WHITE);
  tft.drawLine(88, 48, 98, 38, ILI9341_WHITE);
  tft.drawLine(98, 38, 222, 38, ILI9341_WHITE);
  tft.drawLine(222, 38, 232, 48, ILI9341_WHITE);
  tft.drawLine(232, 48, 232, 132, ILI9341_WHITE);
  tft.drawLine(232, 132, 222, 142, ILI9341_WHITE);
  tft.drawLine(222, 142, 222, 239, ILI9341_WHITE);
  tft.drawLine(98, 142, 222, 142, ILI9341_WHITE);
  tft.drawLine(98, 170, 222, 170, ILI9341_WHITE);
  tft.drawLine(98, 204, 222, 204, ILI9341_WHITE);
  
  // Target ID text
  tft.setCursor(107, 150);
  tft.setTextColor(ILI9341_WHITE);
  tft.setTextSize(2);
  tft.print("TARGET ID");

  // Draw VDI bar
  tft.drawRect(61, 10, 195, 18, ILI9341_WHITE);

  // Draw signal (DEPTH) bar
  tft.drawRect(30, 48, 22, 118, ILI9341_WHITE);

  // Depth text
  tft.setCursor(12, 182);
  tft.setTextColor(ILI9341_WHITE);
  tft.setTextSize(2);
  tft.print("DEPTH");

  // Draw confidence (CONF) bar
  tft.drawRect(267, 48, 22, 118, ILI9341_WHITE);

  // Confidence bar text
  tft.setCursor(254, 182);
  tft.setTextColor(ILI9341_WHITE);
  tft.setTextSize(2);
  tft.print("CONF");
  
  checkBatt();  // Check battery voltage
}

void loop() {                                      
  if (digitalRead(compPin) == LOW) {         // Check for target trigger
    delay(sampleDelay);                      // Delay ADC measurements
    gbAcc = 0;                               // Reset GB accumulator
    discAcc = 0;                             // Reset DISC accumulator
    digitalWrite(debugPin, HIGH);            // Set debug pin high
    for (int i = 1; i <= numSamples; i++) {  // Take a number of GB and DISC readings
      gb = analogRead(gbPin);                // Read GB channel
      disc = analogRead(discPin);            // Read DISC channel
      gbAcc += gb;                           // Accumulate GB readings
      discAcc += disc;                       // Accumulate DISC readings
    }
    digitalWrite(debugPin, LOW);             // Set debug pin low
    gb = gbAcc / numSamples;                 // Average GB readings
    disc = discAcc / numSamples;             // Average DISC readings
    if (gb >= 512) {                         // Re-scale GB reading
      gb -= 512;
    }
    if (disc >= 512) {                       // Re-scale DISC reading
      disc -= 512;
    } else {
      disc = (512 - disc) * -1;    
    }

    angleRad = atan(disc/gb);                // Calculate target phase in radians
    angleDeg = angleRad / TWO_PI * 360;      // Convert to degrees
    vdi = int(angleDeg);                     // Round value for display

    // Overwrite previous VDI bar cursor
    tft.drawChar(vdiBar, 11, 32, ILI9341_GREEN, ILI9341_BLACK, 2);     
    vdiBar = vdi + 90 + 63;  // Calculate new VDI bar position
    if (vdi >= 0) {
      // Display cursor in green when positive
      tft.drawChar(vdiBar, 11, 218, ILI9341_GREEN, ILI9341_BLACK, 2);  
    } else {
      // Display cursor in red when negative
      tft.drawChar(vdiBar, 11, 218, ILI9341_RED, ILI9341_BLACK, 2);    
    }
    
    if (vdi >= 0) {
      targetIDColor = ILI9341_GREEN;  // Display target ID in green when positive
    } else {
      targetIDColor = ILI9341_RED;    // Display target ID in red when negative
      vdi = vdi * -1;                 // Remove negative sign
    }
    targetID = vdi / 10 % 10 + 48;    // Calculate new value for target ID
    // Display first digit of target ID
    tft.drawChar(110, 55, targetID, targetIDColor, ILI9341_BLACK, 8);
    targetID = vdi % 10 + 48;
    // Display second digit for target ID
    tft.drawChar(170, 55, targetID, targetIDColor, ILI9341_BLACK, 8);

    if (targetIDColor == ILI9341_GREEN) {     // Check if target ID is positive
      if (acceptBar >= rejectBar) {           // Check if accept bar is already active
        if (acceptBar != 8) { acceptBar++; }  // Limit extent of accept bar
        // Increase accept bar
        tft.fillRect(269, 106 - acceptBar * 7, 18, acceptBar * 7, ILI9341_GREEN);  
      } else {
        // Limit extent of reject bar
        if (rejectBar != 0) { rejectBar--; }  
        // Decrease reject bar
        tft.fillRect(269, 108 + rejectBar * 7, 18, (8 - rejectBar) * 7, ILI9341_BLACK);  
      }
    } else {
      if (rejectBar >= acceptBar) {           // Check if reject bar is already active
        if (rejectBar != 8) { rejectBar++; }  // Limit extent of reject bar
        tft.fillRect(269, 108, 18, rejectBar * 7, ILI9341_RED);  // Increase reject bar
      } else {
        if (acceptBar != 0) { acceptBar--; }  // Limit extent of accept bar
        // Decrease accept bar
        tft.fillRect(269, 50, 18, (8 - acceptBar) * 7, ILI9341_BLACK);
      }
    }
      
    magnitude = sqrt((gb*gb)+(disc*disc));  // Calculate target signal magnitude
    sig = int(magnitude);                   // Round value for display
    sig -= 22;                              // Offset signal value
    if (sig <= 0) { sig = 1; }              // Limit extent of signal value                         
    sigBar = int(log10(sig) * 10) * 4;      // Find logarithm of signal and scale result
    sigBarY = 163 - sigBar;                 // Offset start of signal bar
    if (sigBar >= prevSigBar) {             // Only update signal bar if necessary
      // Increase signal bar
      tft.fillRect(32, sigBarY, 18, sigBar, ILI9341_WHITE);
    } else {
      // Decrease signal bar
      tft.fillRect(32, prevSigBarY, 18, prevSigBar - sigBar, ILI9341_BLACK);
    }
    prevSigBar = sigBar;                   // Save length of signal bar
    prevSigBarY = sigBarY;                 // Save Y position of signal bar

    dispCount = 0;                         // Reset display count
    delay(100);                            // Wait for 100ms between VDI target indications
  } else {
    tft.setTextSize(2);
    mode = digitalRead(amDisc);            // Check AM/DISC switch
    if (mode != prevMode) {                // Only update display if necessary
      if (mode == LOW) {                   // Test if DISC mode selected
        tft.setCursor(135, 180);
        tft.setTextColor(ILI9341_BLACK);
        tft.print(" AM ");                 // Overwrite AM text
        tft.setCursor(135, 180);
        tft.setTextColor(ILI9341_YELLOW);
        tft.print("DISC");                 // Display mode as DISC
      } else {                             // Otherwise mode must be AM
        tft.setCursor(135, 180);
        tft.setTextColor(ILI9341_BLACK); 
        tft.print("DISC");                 // Overwrite DISC text
        tft.setCursor(135, 180);
        tft.setTextColor(ILI9341_YELLOW);
        tft.print(" AM ");                 // Display mode as AM
      }
    }
    prevMode = mode;                       // Save operating mode
    dispCount++;                           // Increment display counter
    delay(10);                             // Add a 10ms delay to loop
    if (dispCount >= dispCountLimit) {     // Check if display count has reached limit
      dispCount = 0;                       // Clear display count
      checkBatt();                         // Check battery voltage
      tft.drawChar(110, 55, 32, ILI9341_WHITE, ILI9341_BLACK, 8);  // Clear target ID
      tft.drawChar(170, 55, 32, ILI9341_WHITE, ILI9341_BLACK, 8);
      tft.fillRect(64, 11, 191, 16, ILI9341_BLACK);   // Clear VDI bar
      tft.setCursor(30, 207);
      tft.setTextColor(ILI9341_BLACK);
      tft.setTextSize(2);
      tft.fillRect(32, 50, 18, 114, ILI9341_BLACK);   // Clear signal bar
      tft.fillRect(269, 50, 18, 114, ILI9341_BLACK);  // Clear confidence bar
      acceptBar = 0;                                  // Save length of accept bar
      rejectBar = 0;                                  // Save length of reject bar
    }
  }
}
